/*
  optical flow based on the horn schunck method
*/

// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;
float4x4 tV: VIEW;
float4x4 tP: PROJECTION;
float4x4 tWVP: WORLDVIEWPROJECTION;

//textures
texture Tex0 <string uiname="Texture0";>;
sampler Samp0 = sampler_state
{
    Texture   = (Tex0);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

texture Tex1 <string uiname="Texture1";>;
sampler Samp1 = sampler_state
{
    Texture   = (Tex1);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;

// params for optical flow
float2 texSize = float2(640., 480.);
float2 poffset = float2(1., 1.);
float2 scale = float2(1., 1.);
float lambda = 0.;

struct vs2ps
{
    float4 Pos  : POSITION;
    float2 TexCd : TEXCOORD0;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------

vs2ps VS(
    float4 PosO  : POSITION,
    float4 TexCd : TEXCOORD0)
{
    //inititalize all fields of output struct with 0
    vs2ps Out = (vs2ps)0;

    //transform position
    Out.Pos = mul(PosO, tWVP);
    
    //transform texturecoordinates
    Out.TexCd = mul(TexCd, tTex);

    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------

float4 PS(vs2ps In): COLOR
{
    float4 col = tex2D(Samp1, In.TexCd);
    return col;
}

float4 PSOpticalFlowHS(vs2ps In): COLOR
{
    // horn schunck optical flow after pseudo code from http://www.petewarden.com/notes/archives/2005/05/gpu_optical_flo.html  :
    // (1) CurrentDifference = (Frame1[x][y]-Frame0[x][y]);
    // (2) GradientX = (( Frame1[x+1][y]-Frame1[x-1][y]) + (Frame0[x+1][y]-Frame0[x-1][y])) / 2.0;
    // (3) GradientY = ((Frame1[x][y+1]-Frame1[x][y-1]) + (Frame0[x][y+1]-Frame0[x][y-1])) / 2.0;
    // (4) GradientMagnitude = sqrt((GradientX*GradientX)+(GradientY*GradientY))
    // (4 alternative) GradientMagnitude = sqrt((GradientX*GradientX)+(GradientY*GradientY)+Lambda);
    // (5) VelocityX = CurrentDifference*(GradientX/GradientMagnitude);
    // (6) VelocityY = CurrentDifference*(GradientY/GradientMagnitude);

    float2 offset = poffset/texSize;
    //float2 scale=pscale;

    float4 a = tex2D(Samp0, In.TexCd);
    float4 b = tex2D(Samp1, In.TexCd);
    
    float2 texcoord0 = In.TexCd;
    
    float2 nx = float2(offset.x, 0.);
    float2 ny = float2(0., offset.y);

    float4 CurrentDifference = b-a;

    float4 GradientX = ((tex2D(Samp1, texcoord0 + nx) - tex2D(Samp1, texcoord0 - nx)) + (tex2D(Samp0, texcoord0 + nx) - tex2D(Samp0, texcoord0 - nx)))/2.;
    float4 GradientY = ((tex2D(Samp1, texcoord0 + ny) - tex2D(Samp1, texcoord0 - ny)) + (tex2D(Samp0, texcoord0 + ny) - tex2D(Samp0, texcoord0 - ny)))/2.;

    //float4 GradientMagnitude = sqrt( (GradientX*GradientX)+(GradientY*GradientY));
    float4 GradientMagnitude = sqrt( (GradientX*GradientX)+(GradientY*GradientY)+float4(lambda, lambda, lambda, lambda));

    float4 VelocityX = CurrentDifference * (GradientX/GradientMagnitude);
    float4 VelocityY = CurrentDifference * (GradientY/GradientMagnitude);

    float vxd = VelocityX.r; // assumes greyscale
    float2 xout = float2(max(vxd, 0.), abs(min(vxd, 0.)))*scale.x;
    
    float vyd = VelocityY.r; // assumes greyscale
    float2 yout = float2(max(vyd, 0.), abs(min(vyd, 0.)))*scale.y;

    return float4(xout, yout);
}

// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique TSimpleShader
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_1_0 PS();
    }
}

technique TOpticalFlowHS
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_2_0 PSOpticalFlowHS();
    }
}

technique TFixedFunction
{
    pass P0
    {
        //transforms
        WorldTransform[0]   = (tW);
        ViewTransform       = (tV);
        ProjectionTransform = (tP);

        //texturing
        Sampler[0] = (Samp0);
        TextureTransform[0] = (tTex);
        TexCoordIndex[0] = 0;
        TextureTransformFlags[0] = COUNT2;
        //Wrap0 = U;  // useful when mesh is round like a sphere
        
        Lighting       = FALSE;

        //shaders
        VertexShader = NULL;
        PixelShader  = NULL;
    }
}
